/*
spacenavd - a free software replacement driver for 6dof space-mice.
Copyright (C) 2007-2025 John Tsiombikas <nuclear@mutantstargoat.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef KBEMU_H_
#define KBEMU_H_

#include <X11/Xlib.h>
#include <X11/keysym.h>

void kbemu_set_display(Display *dpy);
KeySym kbemu_keysym(const char *str);
const char *kbemu_keyname(KeySym sym);

void send_kbevent(KeySym key, int press);

#endif	/* KBEMU_H_ */
